from collections.abc import Callable
from typing import Any

from aiogram import Dispatcher
from aiogram.dispatcher.event.telegram import TelegramEventObserver

from aiogram3_di.manager import DIManager
from aiogram3_di.middleware import DIMiddleware


def setup_di(
    dispatcher: Dispatcher,
    *,
    allowed_updates: list[str] | None = None,
    dependency_overrides: dict[Callable[..., Any], Callable[..., Any]] | None = None,
) -> DIManager:
    if not isinstance(dispatcher, Dispatcher):
        raise TypeError("dispatcher must be an instance of aiogram.Dispatcher")

    if allowed_updates is not None:
        for allowed_update in allowed_updates:
            if allowed_update not in dispatcher.observers:
                raise ValueError(f"`{allowed_update}` is not a valid allowed update")

    for allowed_update in allowed_updates or dispatcher.resolve_used_update_types():
        observer: TelegramEventObserver = getattr(dispatcher, allowed_update)
        observer.middleware(DIMiddleware())

    dispatcher["di_manager"] = di_manager = DIManager(
        dependency_overrides=(dependency_overrides or {})
    )
    return di_manager



__all__ = (
    "Depends",
    "DIManager",
    "setup_di",
    "__version__",
)

from importlib.metadata import version as _version

from .depends import Depends
from .manager import DIManager
from .setup import setup_di

__version__ = _version("aiogram3-di")



from dataclasses import dataclass, field
from collections.abc import Callable
from typing import Any


@dataclass(frozen=True, slots=True)
class Depends:
    func: Callable[..., Any] | None = None
    use_cache: bool = field(default=True, kw_only=True)



import asyncio
import inspect
from collections.abc import Iterator, Callable
from contextlib import asynccontextmanager
from typing import Any, _AnnotatedAlias, get_args, ContextManager

from .depends import Depends


def get_valid_kwargs(data: dict[str, Any], call: Callable[..., Any]) -> dict[str, Any]:
    valid_params: list[str] = [
        param.name
        for param in inspect.signature(call).parameters.values()
        if param.default == inspect.Parameter.empty
    ]
    return {key: value for key, value in data.items() if key in valid_params}


def get_dependencies(annotations: dict[str, Any]) -> Iterator[tuple[str, Depends, Any]]:
    for annotation_key, annotation_value in annotations.items():
        if isinstance(annotation_value, inspect.Parameter):
            annotation_value = annotation_value.annotation

        if isinstance(annotation_value, _AnnotatedAlias):
            type_annotation, dependency = get_args(annotation_value)[:2]
            if isinstance(dependency, Depends):
                call = dependency.func or type_annotation
                if parameters := inspect.signature(call).parameters:
                    yield from get_dependencies(dict(parameters))
                yield annotation_key, dependency, type_annotation


@asynccontextmanager
async def contextmanager_in_threadpool(cm: ContextManager):
    try:
        yield await asyncio.to_thread(cm.__enter__)
    except Exception as e:
        ok = bool(await asyncio.to_thread(cm.__exit__, type(e), e, None))
        if not ok:
            raise e
    else:
        await asyncio.to_thread(cm.__exit__, None, None, None)


def is_coroutine_callable(call: Callable[..., Any]) -> bool:
    if inspect.isroutine(call):
        return inspect.iscoroutinefunction(call)
    if inspect.isclass(call):
        return False
    dunder_call = getattr(call, "__call__", None)
    return inspect.iscoroutinefunction(dunder_call)


def is_async_gen_callable(call: Callable[..., Any]) -> bool:
    if inspect.isasyncgenfunction(call):
        return True
    dunder_call = getattr(call, "__call__", None)
    return inspect.isasyncgenfunction(dunder_call)


def is_gen_callable(call: Callable[..., Any]) -> bool:
    if inspect.isgeneratorfunction(call):
        return True
    dunder_call = getattr(call, "__call__", None)
    return inspect.isgeneratorfunction(dunder_call)



from collections.abc import Callable, Awaitable
from contextlib import AsyncExitStack
from typing import Any

from aiogram import BaseMiddleware
from aiogram.dispatcher.flags import get_flag
from aiogram.types import TelegramObject

from .resolver import DependenciesResolver


class DIMiddleware(BaseMiddleware):
    async def __call__(
        self,
        handler: Callable[[TelegramObject, dict[str, Any]], Awaitable[Any]],
        event: TelegramObject,
        data: dict[str, Any],
    ) -> Any:
        handler_dependencies = tuple(get_flag(data, "dependencies", default=()))

        async with AsyncExitStack() as stack:
            resolver = DependenciesResolver(
                stack,
                handler_dependencies=handler_dependencies,
                event=event,
                middleware_data=data.copy(),
            )
            data = await resolver.resolve()
            return await handler(event, data)



import asyncio
import inspect
from collections.abc import Callable
from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from typing import Any, TypeAlias, _AnnotatedAlias

from aiogram.dispatcher.event.handler import HandlerObject
from aiogram.types import TelegramObject

from .depends import Depends
from .manager import DIManager
from .utils import (
    get_dependencies,
    get_valid_kwargs,
    is_async_gen_callable,
    is_gen_callable,
    contextmanager_in_threadpool,
    is_coroutine_callable,
)

ParamData: TypeAlias = dict[tuple[int, str], Any]

CACHE_MISSING = object()


class DependenciesResolver:
    __slots__ = (
        "_stack",
        "_handler_dependencies",
        "_event",
        "_middleware_data",
        "_param_data",
        "_cache",
    )

    def __init__(
        self,
        stack: AsyncExitStack,
        *,
        handler_dependencies: tuple[Depends, ...],
        event: TelegramObject,
        middleware_data: dict[str, Any],
    ) -> None:
        self._stack = stack
        self._handler_dependencies = handler_dependencies
        self._event = event
        self._middleware_data = middleware_data
        self._param_data: ParamData = {}
        self._cache: dict[int, Any] = {}

    async def resolve(self) -> dict[str, Any]:
        for handler_dependency in self._handler_dependencies:
            await self._resolve_handler_dependency(handler_dependency)

        handler: HandlerObject = self._middleware_data["handler"]
        handler_annotations = inspect.get_annotations(handler.callback)

        for param_name, dependency, type_annotation in get_dependencies(
            handler_annotations
        ):
            await self._process_dependency(param_name, dependency, type_annotation)

        return self._middleware_data | self._extract_params(handler_annotations)

    def _extract_params(self, annotations: dict[str, Any]) -> dict[str, Any]:
        dependencies: list[tuple[int, str]] = []

        for param_name, param_value in annotations.items():
            if isinstance(param_value, _AnnotatedAlias) and isinstance(
                param_value.__metadata__[0], Depends
            ):
                dependency: Depends = param_value.__metadata__[0]
                type_annotation: Any = param_value.__origin__
                call = self._get_call(dependency, type_annotation)
                dependencies.append((hash(call), param_name))

        return {
            param_name: self._param_data[(func_hash, param_name)]
            for func_hash, param_name in dependencies
        }

    async def _resolve_handler_dependency(self, dependency: Depends) -> None:
        for param_name, sub_dependency, type_annotation in get_dependencies(
            inspect.get_annotations(dependency.func)
        ):
            await self._process_dependency(param_name, sub_dependency, type_annotation)

        call = self._get_call(dependency, None)
        await self.__process_dependency(call, dependency)

    async def _process_dependency(
        self, param_name: str, dependency: Depends, type_annotation: Any
    ) -> None:
        call = self._get_call(dependency, type_annotation)
        result = await self.__process_dependency(call, dependency)
        self._param_data[(hash(call)), param_name] = result

    def _get_call(
        self, dependency: Depends, type_annotation: Any
    ) -> Callable[..., Any]:
        original_call = dependency.func or type_annotation
        di_manager: DIManager = self._middleware_data["di_manager"]
        call = di_manager.dependency_overrides.get(original_call, original_call)
        return call

    async def __process_dependency(
        self, call: Callable[..., Any], dependency: Depends
    ) -> Any:
        cached_value = self._cache.get(hash(call), CACHE_MISSING)

        if cached_value is not CACHE_MISSING and dependency.use_cache:
            return cached_value

        params = self._extract_params(inspect.get_annotations(call))
        values = get_valid_kwargs(
            (self._middleware_data | {"event": self._event} | params), call
        )

        if is_async_gen_callable(call):
            cm = asynccontextmanager(call)(**values)
            result = await self._stack.enter_async_context(cm)
        elif is_gen_callable(call):
            cm = contextmanager_in_threadpool(contextmanager(call)(**values))
            result = await self._stack.enter_async_context(cm)
        elif is_coroutine_callable(call):
            result = await call(**values)
        else:
            result = await asyncio.to_thread(call, **values)
        self._cache.setdefault(hash(call), result)
        return result



from collections.abc import Callable
from typing import Any


class DIManager:
    __slots__ = "_dependency_overrides"

    def __init__(
        self,
        *,
        dependency_overrides: dict[Callable[..., Any], Callable[..., Any]],
    ) -> None:
        self._dependency_overrides = dependency_overrides

    @property
    def dependency_overrides(self) -> dict[Callable[..., Any], Callable[..., Any]]:
        return self._dependency_overrides

    @dependency_overrides.setter
    def dependency_overrides(
        self, value: dict[Callable[..., Any], Callable[..., Any]]
    ) -> None:
        self._dependency_overrides = value
